****************************************************************************
*  User modifiable functions used in NT2/CA3 for handling serial           *
*  communication. Note, FlagShip support 16 ports instead of 4 in CA3,     *
*  but usually there are only 4 ports initialized by the system.           *
*                                                                          *
*  If you need to change this file, copy it into your local directory,     *
*  make changes and recompile it by using                                  *
*     FlagShip fs2_com.prg -c -m -na                                       *
*  and then link it with your application                                  *
*     FlagShip myprog*.prg fs2_com.o -lFS2                                 *
*                                                                          *
*  Note: When FS_SET("devel", .T.) is active, you will receive an          *
*        additional iformation on unsupported functions.                   *
*                                                                          *
*  Note: if you want to debug the C inline parts by the C debugger         *
*        e.g. with gdb, cv, xldb etc, your should:                         *
*        cd <my_working_dir>                                               *
*        cp /usr/FS2src/fs2_com.prg .                                      *
*        cp /etc/FSconfig .                                                *
*          change cc -> cc -g   in CCNAME: line                            *
*          remove the -s switch in POSTOPTIONS: line                       *
*        FlagShip fs2_com.prg [-DTESTPRINT] -c -m -na -nL -nl              *
*        FlagShip myprog*.prg fs2_com.o -lFS2 [/usr/FS2src/bb*.o]          *
*        gdb a.out                                                         *
*        list _bb_fs2_termio  ...etc  to debug the fs2_termio() function   *
*        br 3218              ...etc, see help                             *
*        r   [c,n,s...]                                                    *
*        quit                                                              *
*                                                                          *
*                           (c) 1992-99 multisoft Datentechnik GmbH Munich *
****************************************************************************
* Differences and extentions to CA3 (see text in the FN header below):
*
* com_break (nPort, nTime) -> .F.
* com_crc   (cString, [nStart], [nPoly]) -> nCRC
* com_open  (nPort, [nSize], [nOut], [lTrigg], [cDevice]) -> lOk
* com_cts   (nPort) -> .T.
* com_dcd   (nPort) -> .T.
* com_dsr   (nPort) -> .T.
* com_dtr   (nPort) -> .T.
* com_errchr(nPort, ncChar) -> .F.
* com_error (nPort) -> nError
* com_event (nPort, nMode)  -> -1
* com_getio (nPort) -> nFileHd
* com_getirq(nPort) -> nFileHd
* com_init  (nPort, [nBaud], [cParity], [nDataLng], [nStop], ;
*            [lCtsRts], [lLocal], [lCanon]) -> lOk
* com_init2 (nPort, [nBaud], [cParity], [nDataLng], [nStop]) -> lOk
* com_key   (nPort, nKey1, nKey2) -> .F.
* com_lsr   (nPort) -> nError
* com_mcr   (nPort) -> 0
* com_msr   (nPort) -> 0
* com_num   ()      -> nPorts == MAXPORTS
* com_remote(nPort, [ncChar]) -> .F.
* com_ring  (nPort) -> .T.
* com_rts   (nPort, [lNew]) -> .T.
* com_setio (nPort, [ncPortAdr])  -> .F.
* com_setirq(nPort, [ncPortiIrq]) -> .F.
* com_skey  (nPort, [ncKey1], [ncKey2]) -> .F.
* com_soft  (nPort, [lNew], [cXon], [cXoff]) -> .F.
* com_soft_s(nPort) -> .F.
****************************************************************************

#Cinline
#include "FSopenc.h"
#include <malloc.h>
#include <string.h>

/* example 
 * test helper to print bits in zeroinsert() or zeroremove()
 * active and available only when compiled with -DTESTPRINT
 */
#ifdef TESTPRINT
void printbuff(unsigned char cc, unsigned char *buff, int size)
{
   int ii;
   unsigned char *ptr;

   printf("\nChar=%.2X Buff=", cc);
   for (ptr=buff; ptr < buff + size; ptr++) {
     for (ii=7; ii >= 0; ii--)
        printf("%c", ((*ptr & (1 << ii)) ? '1' : '0'));
     printf(" ");
   }
}
#endif
#endCinline

#include "set.fh"

#define MAXPORTS  16
#define BUFFSIZE  1024

static anFileHd[MAXPORTS]            // file handle for COM1 ... COM16
static acBuffer[MAXPORTS]            // for com_count(), com_read()
static anBufLen[MAXPORTS]            // valid chars in cBuffRead    (0, 1..n)
static anBufPos[MAXPORTS]            // current pointer in cBuffRead (0..n-1)
static anLastEr[MAXPORTS]            // last r/w error on this port (if > 0)
static asStatus[MAXPORTS]            // saved port status on com_open()

static MSG_PORT0   := "Port manipulation not possible"
static MSG_PORT    := MSG_PORT0 + ", returning "
static MSG_COMINIT := "Port manipulation not possible, use;" + ;
                      "com_init() or com_init2(). Returning "

*****************************************************************************
* com_break(nPort, nTime) -> lOk
*
* Creates break on a transmission line
*
* THIS FUNCTION IS NOT APPLICABLE ON UNIX
* Default value .F. is returned
*----------------------------------------------------------------------------
FUNCTION com_break(nPort, nTime)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_close(nPort) -> lOk
*
* Clears the receiving buffer and closes the com port
*----------------------------------------------------------------------------
FUNCTION com_close(nPort)
local ok
if !com_isopen(nPort)
   return .F.
endif
if valtype(asStatus[nPort]) != "U"
   fs2_termio(-2, nPort, @asStatus[nPort])  // restore old status
endif   
ok := FCLOSE(anFileHd[nPort])
if ok
   anFileHd[nPort] := 0
   anBufLen[nPort] := 0
   anBufPos[nPort] := 0
   anLastEr[nPort] := 0
   acBuffer[nPort] := NIL
   asStatus[nPort] := NIL
endif
return ok

*****************************************************************************
* com_count(nPort) -> nChar
*
* Counts the number of characters in the input buffer
*
* nPort: 1 to 16 for COM1 to COM16
* nChar: >0 data in the device buffer are available,
*        =0 no data in COM buffer available 
*        -1 error (port not open)
*
* This function is tightly related to com_read(). It reads
* up to 1k characters from the device into internal buffer to
* be subsequentely be read via com_read().
*----------------------------------------------------------------------------
FUNCTION com_count(nPort)
if !com_isopen(nPort)
   return -1 
endif
if anBufLen[nPort] > 0 .and. anBufPos[nPort] < anBufLen[nPort]
   return anBufLen[nPort] - anBufPos[nPort]
endif
anBufLen[nPort] := FREAD(anFileHd[nPort], @acBuffer[nPort], len(acBuffer[nPort]))
anBufPos[nPort] := 0
anLastEr[nPort] := FERROR()
return anBufLen[nPort]

*****************************************************************************
* com_crc(cString, [nStart], [nPoly]) -> nCRC
*
* Computes CRC value for the string <cString>
* 
* cString : the string for which the CRC value is computed
* nStart  : optional, value added to resulting nCRC
* nPoy    : optional, used Polynomial (any int value > 0)
*           once set, the polynomial remains active until a new nPoly is set
* nCRC    : resulting CRC value
*
* XMODEM (default)      nStart = 0,     nPoly = 40961 
* X25                   nStart = 0,     nPoly =  4129
* CCITT                 nStart = 65535, nPoly =  4129
* ARC                   nStart = 0,     nPoly = 32773
*
* CA3: Note, the NT3/CA3 algorithm is compatible only to istelf!
* FS2: FlagShip follows the international CRC16 standard.
*----------------------------------------------------------------------------
* function com_crc(cString, nStart, nPoly)
* this function is available in bbcrc16.o

*****************************************************************************
* com_cts(nPort) -> lActive
*
* CA3: Queries the Clear To Send (CTS) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .T. is returned here
*      To set the RTS/CTS auto handshake, use com_init()
*----------------------------------------------------------------------------
FUNCTION com_cts(nPort)
fs2_devWarn(MSG_COMINIT + ".F.")
return .T.

*****************************************************************************
* com_dcd(nPort) -> lActive
*
* CA3: Queries the Data Carrier Detect (DCD) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .T. is returned here
*      To set the RTS/CTS auto handshake, use com_init()
*----------------------------------------------------------------------------
FUNCTION com_dcd(nPort)
fs2_devWarn(MSG_COMINIT + ".T.")
return .T.

*****************************************************************************
* com_dsr(nPort) -> lActive
*
* CA3: Queries the Data Set Ready (DSR) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .T. is returned here
*      To set the RTS/CTS auto handshake, use com_init()
*----------------------------------------------------------------------------
FUNCTION com_dsr(nPort)
fs2_devWarn(MSG_COMINIT + ".T.")
return .T.

*****************************************************************************
* com_dtr(nPort, [lNew]) -> lActive
*
* CA3: Queries or set the Data Terminal Ready (DTR) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .T. is returned here
*      To set the RTS/CTS auto handshake, use com_init()
*----------------------------------------------------------------------------
FUNCTION com_dtr(nPort, lNew)
fs2_devWarn(MSG_COMINIT + ".T.")
return .T.

*****************************************************************************
* com_errchr(nPort, [ncChar]) -> lChanged
*
* CA3: Defines replacement for a "trash" character
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_errchr(nPort, ncChar)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_error(nPort) -> nError
*
* FS2: Reports the last r/w error, if any
*      Not applicable for com_open() and com_close(), where  FERROR() 
*      should be used instead.
* CA3: not available, use com_lsr
*
* nError = last error (see the FERROR function) on this port
*----------------------------------------------------------------------------
FUNCTION com_error(nPort)
if !com_isopen(nPort)
   return  0
endif
return anLastEr[nPort]

*****************************************************************************
* com_event(nPort, nMode) -> nRet
*
* CA3: Event triggering of a port
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value -1 is returned here
*----------------------------------------------------------------------------
FUNCTION com_event(nPort, nMode)
fs2_devWarn(MSG_PORT + "-1")
return -1

*****************************************************************************
* com_flush(nPort) -> lOk
*
* Flushes the internal buffer of com_count()
* However, it does not flush the rest of incomming chars on the serial line
*----------------------------------------------------------------------------
FUNCTION com_flush(nPort)
if !com_isopen(nPort)
   return .F.
endif
anBufLen[nPort] := 0
anBufPos[nPort] := 0
return .T.

*****************************************************************************
* com_getio(nPort) -> nFileHd
*
* FS2: Determines the file handle used for an open <nPort>
* CA3: returns the I/O port address instead
*----------------------------------------------------------------------------
FUNCTION com_getio(nPort)
if !com_isopen(nPort)
   return -1
endif
return anFileHd[nPort]

*****************************************************************************
* com_getirq(nPort) -> nFileHd
*
* FS2: Determines the file handle used for an open <nPort>
* CA3: returns the IRQ address instead
*----------------------------------------------------------------------------
FUNCTION com_getirq(nPort)
if !com_isopen(nPort)
   return -1
endif
return anFileHd[nPort]

*****************************************************************************
* com_hard(nPort, [lOn], [lDtr]) -> lOldRts
*
* CA3: Sets the DTR/DSR ort RTS/CTS handshake
* FS2: only RTS/CTS handshake (lDtr == .F.) is supported, port must be open!
*----------------------------------------------------------------------------
FUNCTION com_hard(nPort, lOn) 
local lRet := .F.
if com_isopen(nPort)
   lRet := fs2_termio(1, nPort)
   if valtype(lOn) == "L"
      fs2_termio(1, nPort, lOn)
   endif
endif
return lRet

*****************************************************************************
* com_init(nPort, [nBaud], [cParity], [nDataLng], [nStop], ;
*          [lCtsRts], [lLocal], [lCanon]) -> lOk
*
* Sets the serial port attributtes
*
* nPort   : 1 to 16, corresponding to com_open()
* nBaud   : 300, 1200, 2400, 4800, (9600), 19200, 38400, 57600, 115200, 23040
* cParity : ("N[o]") , "E[ven]", "O[dd]"
* nDataLng: 5, 6, 7 or (8)
* nStop   : 0, (1), 2
* lCtsRts : enable (.T.) or disable [.F.] CTS/RTS 
* lLocal  : local  (.T.) or dial-up [.F.] line 
* lCanon  : canonical [.T.] inut processing = editeable or non-canoncal (.F.)
*
* If the optional parameter are not given or are NIL, the default values
* given above in parentheses () are used. So when all optional parameters 
* are omited, the default setting   
*     com_init(nPort, 9600, "N", 8, 1, .T., .T., .F.)   
* is used.
*
* CA3: The default speed is 300 baud instead 9600 in FS2. In CA3, only
*      the first five parameters are supported.
*----------------------------------------------------------------------------
FUNCTION com_init(nPort, nBaud, cParity, nDataLng, nStop, lCtsRts, lLocal, lCanon)
local lRet   := .F.

if com_isopen(nPort)
   nBaud    := if(valtype(nBaud) $ "NI", nBaud, 9600)
   cParity  := if(valtype(cParity) == "C", upper(left(cParity,1)), "N")
   cParity  := if(cParity $ "NEO", cParity, "N")
   nDataLng := if(valtype(nDataLng) $ "NI", nDataLng, 8)
   nDataLng := if(ascan({5,6,7,8}, nDataLng) > 0, nDataLng, 8)  
   nStop    := if(valtype(nStop) $ "NI", nStop, 1)
   nStop    := if(ascan({0,1,2}, nStop) > 0, nStop, 1)  
   lCtsRts  := if(valtype(lCtsRts) == "L", lCtsRts, .T.)
   lLocal   := if(valtype(lLocal)  == "L", lLocal,  .T.)
   lCanon   := if(valtype(lCanon)  == "L", lCanon,  .F.)
                 
   lRet := fs2_termio(0, nPort, nBaud, nDataLng, cParity, nStop, ;
                      lCtsRts, lLocal, lCanon)
endif
return lRet 

*****************************************************************************
* com_init2(nPort, [nBaud], [cParity], [nDataLng], [nStop]) -> lOk
*
* Sets the serial port attributtes
*
* nPort   : 1 to 16, corresponding to com_open()
* nBaud   : 300, 1200, 2400, 4800, (9600), 19200, 38400, 57600, 115200, 23040
* cParity : ("N[o]") , "E[ven]", "O[dd]"
* nDataLng: 5, 6, 7 or (8)
* nStop   : 0, (1), 2
*
* As opposed to com_init() where all the port parameters are initialized,
* com_init2() sets only the specified parameters, i.e. not NIL. 
*
* CA3: not available
*----------------------------------------------------------------------------
FUNCTION com_init2(nPort, nBaud, cParity, nDataLng, nStop)
local lRet   := .T.

if com_isopen(nPort)
   if valtype(nBaud) == "N" .and. !empty(nBaud)
       fs2_devWarn("Setting general serial speed. To initialize;" + ;
                   "specific port, use com_init() instead.")
       lRet  := fs2_termio(2, nPort, nBaud) == nBaud
   endif
   if valtype(cParity) == "C" .and. !empty(cParity) .and. lRet
       lRet  := fs2_termio(3, nPort, if(upper(left(cParity,1)) == "N", 0, 1))
   endif
   if valtype(nDataLng) == "N" .and. !empty(nDataLng) .and. lRet
      lRet  := fs2_termio(4, nPort, nDataLng) 
   endif
   if valtype(nStop) == "N" .and. lRet
      lRet  := fs2_termio(5, nPort, nStop)
   endif
endif
return lRet 

*****************************************************************************
* com_isopen(nPort) -> lOk
*
* Checks if the port is already open
*
* CA3: This function is not available in CA3
*----------------------------------------------------------------------------
FUNCTION com_isopen(nPort)
if valtype(nPort) $ "NI" .and. ;
   nPort > 0 .and. nPort <= len(anFileHd) .and. ;
   valtype(anFileHd[nPort]) $ "NI" .and. anFileHd[nPort] > 0 .and. ;
   valtype(acBuffer[nPort]) == "C" .and. len(acBuffer[nPort]) > 0
   return .T.
endif
return .F.

*****************************************************************************
* com_key(nPort, [nKey1], [nKey2]) -> lTrap
*
* CA3: Monitors port triggering
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_key(nPort, nKey1, nKey2)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_lsr(nPort) -> nError
*
* FS2: Reports the last error, if any
* CA3: Reports the Status Register (LSR)
*
* nError = last error (see the FERROR function) on this port
*----------------------------------------------------------------------------
FUNCTION com_lsr(nPort)
return com_error(nPort)

*****************************************************************************
* com_mcr(nPort) -> nRet
*
* CA3: Reads or sets the Modem Control Register (MCR)
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value 0 is returned here
*----------------------------------------------------------------------------
FUNCTION com_mcr(nPort)
fs2_devWarn(MSG_PORT + "0")
return  0

*****************************************************************************
* com_msr(nPort) -> nRet
*
* CA3: Reads or sets the Modem Status Register (MSR)
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value 0 is returned here
*----------------------------------------------------------------------------
FUNCTION com_msr(nPort)
fs2_devWarn(MSG_PORT + "0")
return  0

*****************************************************************************
* com_num() -> nPorts
*
* FS2: Reports the number of supported ports (default = 16)
* CA3: Reports the number of highest available serial port
*----------------------------------------------------------------------------
FUNCTION com_num()
return len(anFileHd)

*****************************************************************************
* com_open(nPort, [nSize], [nOut], [lTrigg], [cDevice]) -> lOk
*
* Opens the port and initializes the buffer
* When the special device <cDevice> is not given, the corresponding
* serial device to nPort will be opened. Since there ae two standards,
* we first try to open /dev/cua<n> and then /dev/ttyS<n> where <n>
* corresponds to <nPort> -1, i.e. 3 stay for COM3 or /dev/ttyS2.
*
* nPort  : 1..16 for port COM1 .. COM16
* nSize  : optional, size of the receiving buffer, default = BUFFSIZE
* nOut   : Output buffer for background. Not used by FS2, ignored
* lTrig  : Triggering by interrupt. Not applicable in FS2, ignored.
* cDevice: optional, device if other than /dev/cua<n> or /dev/ttyS<n>
* lOk    : .T. if successfully opened,
*          .F. on error (port is open, or no file handlers available,
*              could not open, permissions etc.). You may check FERROR() 
*              after !com_open() for the error reason.
*
* CA3: Only 4 parameters are supported
* FS2: Note, you may use the com_*() functions also on any other devices, 
*      not only serial, by using the 5th parameter, e.g. for LPT1: or console
*      com_open(5, , , , "/dev/lp1") or com_open(10, , , , "/dev/tty3") etc.
*      The standard FOPEN() function opens the port in non-blocking mode.
* 
* DOS driver         Unix device
* COM1:              /dev/cua0   or  /dev/ttyS0   or user defined
* COM2:              /dev/cua1   or  /dev/ttyS1   or user defined
*  :
* COM16:             /dev/cua15  or  /dev/ttyS15  or user defined
*----------------------------------------------------------------------------
FUNCTION com_open(nPort, nSize, dum3, dum4, cDevice)
if com_isopen(nPort)
   return .F.
endif
if !(valtype(nPort) $ "NI") .or. ;
   (nPort <= 0 .or. nPort > len(anFileHd)) 
   return .F.
endif

nSize := if(valtype(nSize) $ "NI" .and. nSize > 0, nSize, BUFFSIZE)

anLastEr[nPort] := 0
anBufLen[nPort] := 0
anBufPos[nPort] := 0
acBuffer[nPort] := space(nSize)
if valtype(cDevice) == "C" .and. !empty(cDevice)
   anFileHd[nPort] := FOPEN(alltrim(cDevice), 2)
else
   // Unix default devices are  /dev/cua<n>  -or-  /dev/ttyS<n>
   anFileHd[nPort] := FOPEN("/dev/cua" + ltrim(nPort -1), 2)
   if anFileHd[nPort] < 1
      anFileHd[nPort] := FOPEN("/dev/ttyS" + ltrim(nPort -1), 2)
   endif
endif
if anFileHd[nPort] > 0
   fs2_termio(-1, nPort, @asStatus[nPort])
   return .T.
endif   
return .F.

*****************************************************************************
* com_read(nPort, [nSize], [lHold]) -> cString
*
* CA3: Reads characters from the internal buffer.
* FS2: Reads characters from the internal buffer.  Additionally, when <nSize>
*      is larger than the rest of bytes in the internal buffer, the remaining
*      data are also repeatly read from the device, so the 
*      com_count() + com_read() loop sequence is not required in your prg. 
*      But: when lHold is .T., only the rest of the internal buffer is 
*      returned, no additional data are read from the device.
*
* nPort  : corresponds to the <nPort> from com_open()
* nSize  : number of characters to be read. The default is the rest 
*          of internal buffer; if this is empty, a new set is read
* lHold  : .F. (default) remove read characters from the buffer
*          .T. hold characters in the buffer
* cString: receiving characters of the size <nSize>, or less if no 
*          additional data are available.
*----------------------------------------------------------------------------
FUNCTION com_read(nPort, nSize, lHold)
local nAvail, nLen, lMustRead := .F., cRet := ""
if !com_isopen(nPort)
   return cRet
endif
nAvail := com_count(nPort)     // read into buffer if required
if nAvail <= 0
   return cRet
endif

if !(valtype(nSize) $ "NI") .or. nSize <= 0
   nSize := nAvail
endif
nLen := nSize

if nSize > nAvail
   lMustRead := .T.
   nLen := nAvail
endif
cRet := substr(acBuffer[nPort], anBufPos[nPort] +1, nLen)
if valtype(lHold) == "L" .and. lHold
   return cRet                         // don't remove
endif

anBufPos[nPort] := min(anBufPos[nPort] + nLen, anBufLen[nPort])

while lMustRead 
   nSize  := max(nSize - nAvail, 0)
   nAvail := com_count(nPort)     // read into buffer if required

   lMustRead := .F.
   if nAvail > 0 .and. nSize > 0
      nLen := min(nSize, nAvail)
      cRet += substr(acBuffer[nPort], anBufPos[nPort] +1, nLen)
      lMustRead := nSize > nAvail
      anBufPos[nPort] := anBufPos[nPort] + nLen
   endif
enddo
return cRet 

*****************************************************************************
* com_remote(nPort, [ncChar]) -> lAct
*
* CA3: Determines the clear char for the receiving buffer
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_remote(nPort, ncChar)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_ring(nPort) -> lAct
*
* CA3: Queries the ring line
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .T. is returned here
*----------------------------------------------------------------------------
FUNCTION com_ring(nPort)
fs2_devWarn(MSG_PORT + ".T.")
return .T.

*****************************************************************************
* com_rts(nPort, [lNew]) -> lAct
*
* CA3: Queries or sets the Request To Send (RTS)
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .T. is returned here
*----------------------------------------------------------------------------
FUNCTION com_rts(nPort, lNew)
fs2_devWarn(MSG_PORT + ".T.")
return .T.

*****************************************************************************
* com_scount(nPort) -> nChar
*
* Counts the number of chars in the internal buffer
*----------------------------------------------------------------------------
FUNCTION com_scount(nPort)
if !com_isopen(nPort)
   return -1
endif
return anBufLen[nPort] - anBufPos[nPort]

*****************************************************************************
* com_send(nPort, ncString) -> nFailed
*
* nPort   : corresponds to the <nPort> from com_open()
* ncString: a string or an ASCII value of one characted to be trensmitted
* nFailed : number of charecters not transmitted. The ok case is 0, 
*          on parameter error or if the port is not open, -1 is returned.
*----------------------------------------------------------------------------
FUNCTION com_send(nPort, ncString)
local cBuff, ii
if !com_isopen(nPort)
   return -1
endif
if valtype(ncString) == "C"
   cBuff := ncString
elseif valtype(ncString) == "N" .and. ncString >= 0 .and. ncString <= 255
   cBuff := chr(ncString)
else
   return -1
endif
ii := FWRITE(anFileHd[nPort], cBuff)
anLastEr[nPort] := FERROR() 
return len(cBuff) - ii

*****************************************************************************
* com_setio(nPort, [ncPortAdr]) -> lAct
*
* CA3: Changes the I/O address for a port
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_setio(nPort, ncPortAdr)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_setirq(nPort, [ncPortiIrq]) -> lAct
*
* CA3: Changes the IRQ for a port
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_setirq(nPort, ncPortIrq)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_sflush(nPort) -> lOk
*
* Flushes the internal buffer of com_count()
* However, it does not flush the rest of incomming chars on the serial line
*----------------------------------------------------------------------------
FUNCTION com_sflush(nPort)
return com_flush(nPort)

*****************************************************************************
* com_skey(nPort, [ncKey1], [ncKey2]) -> lOk
*
* CA3: Monitor the port using key traps
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*      You may simply use cRead := com_read() ; KEYBOARD (cRead)
*----------------------------------------------------------------------------
FUNCTION com_skey(nPort, ncKey1, ncKey2)
fs2_devWarn(MSG_PORT0 + ";use cRead:=com_read() + KEYBOARD (cRead);Returning .F.")
return .F.

*****************************************************************************
* com_smode(nPort) -> lOk
*
* Determines the current transmission status
*
* lOk : 0 = port not open
*       1 = buffer is empty
*       2..8 = not applicable in FS2
*      -1 = buffer is available for com_read()
*----------------------------------------------------------------------------
FUNCTION com_smode(nPort)
if !com_isopen(nPort)
   return  0
endif
return if(anBufLen[nPort] - anBufPos[nPort] > 0, -1, 1)

*****************************************************************************
* com_soft(nPort, [lNew], [cXon], [cXoff]) -> lOk
*
* Queries or sets new XON/XOFF handshake or character
* FS2: settong of different [cXon], [cXoff] are not supported
*----------------------------------------------------------------------------
FUNCTION com_soft(nPort, lNew, cXon, cXoff)
if !com_isopen(nPort)
   return fs2_termio(6, nPort, lNew)  
endif
return .F.

*****************************************************************************
* com_soft_r(nPort, [lReset]) -> lXoff
*
* CA3: Tests if an XOFF character has been received, reset Xoff flag
* FS2: Checks if Ctrl-Q (XOFF) is available in receiving buffer
*----------------------------------------------------------------------------
FUNCTION com_soft_r(nPort, ldummy)
if !com_isopen(nPort) .or. anBufLen[nPort] <= 0 .or. ;
   anBufPos[nPort] >= anBufLen[nPort]
   return  .F.
endif
return chr(17) $ substr(acBuffer[nPort], anBufPos[nPort] +1)

*****************************************************************************
* com_soft_s(nPort) -> lXoff
*
* CA3: Tests if an XOFF character has been sent
* FS2: Not applicable. Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_soft_s(nPort)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* xmoblock(cString, nBlockNo, [lCRC], [nMode]) -> cXmodBlock
*
* Generates a block for XMODEM transmission
*
* cString : char string of 128 or 1024 length, depending on <nMode>.
*           Trailing chars are cut off, missing chars filled with binary 0.
* nBlockNo: The related XMODEM block number (BN). starting with 0
* lCRC    : .T. = the modem block contain checksum (CS)
*           .F. = the modem block contain 16-bit CRC (default)
* nMode   : 1 = the XMODEM block contains 
*                 chr(1) + BM + ~BM + 128chars + (CS or CRC)
*           2 = the XMODEM block contains
*                 chr(2) + BM + ~BM + 1024chars + (CS or CRC)
* cXmodBlock = the returned XMODEM block
*----------------------------------------------------------------------------
FUNCTION xmoblock(cString, nBlockNo, lCRC, nMode)
local cBuff, iCheck, cCheck
local b1, b2 as intvar

lCRC  := if (valtype(lCRC) == "L", lCRC, .F.)
nMode := if (valtype(nMode) $ "NI", nMode, 1)

cBuff := padr(cString, if(nMode == 1, 128, 1024), chr(0))
b1 := nBlockNo % 256
b2 := numand (numnot(b1), 255)
if lCRC
   cCheck := chr(numlow(asciisum(cBuff)))
else
   iCheck:= crc16(cBuff, 0, 40961)
   cCheck := chr(numlow(iCheck)) + chr(numhigh(iCheck))
endif
return  if(nMode == 1, chr(1), chr(2)) + chr(b1) + chr(b2) + cBuff + cCheck 

*****************************************************************************
* xmocheck(cString, [lCRC]) -> nBlockNo
*
* Tests a received XMODEM block
*
* cString : char string including a valid XMODEM block, see xmoblock()
* lCRC    : .T. = the modem block contain checksum (CS)
*           .F. = the modem block contain 16-bit CRC (default)
* nBlockNo: The received XMODEM block number (BN) (0..255)
*----------------------------------------------------------------------------
FUNCTION xmocheck(cString, lCRC)
local iCheck, xCheck, ok, ii, nMode, nSize
local b1, b2 as intvar

lCRC  := if (valtype(lCRC) == "L", lCRC, .F.)
nMode := strpeek(@cString, 1)
b1    := strpeek(@cString, 2)
b2    := strpeek(@cString, 3)

ok := nMode == 1 .or. nMode == 2
ok := ok .and. len(cString) >= if(nMode == 1, 131, 1027)
ok := ok .and. b2 == numand (numnot(b1), 255)
if !ok
   return -1
endif
nSize := if(nMode == 1, 128, 1024)
if lCRC
   xCheck := strpeek(@cString, nSize + 3)
   iCheck := numlow(asciisum(substr(cString,4,nSize)))
else
   if len(cString) < nSize + 4
      return -1
   endif
   ii := nSize +4
   xCheck := strpeek(@cString, ii--) * 256 
   xCheck += strpeek(@cString, ii)
   iCheck := crc16(substr(cString,4,nSize), 0, 40961)
endif
if xCheck != iCheck
   return -1
endif
return b1


*****************************************************************************
* zeroinsert(cString) -> cNewStr
*
* Inserts a 0-bit after every fifth 1-bit to allow ISDN or X.25 transfer
*
* cString : char string
* cNewStr : <cString> with inserted 0-bits, the new length may be 20% larger
*----------------------------------------------------------------------------
FUNCTION zeroinsert(cString)
local cRet := ""
if valtype(cString) == "C" .and. len(cString) > 0
#Cinline
{
   int sizeIn, sizeOut, sizeInit, added = 0, move = 0, byteNo = 0;
   unsigned char *cIn, *cOut, *ptr, *buff;
   unsigned char cc, pp, dd, rr;
   
   sizeIn   = VAR_ISLEN(VAR_NAME_LOCPAR(cstring)) -1;
   sizeInit = sizeIn * 1.2 + 3;
   cIn  = VAR_CHR(VAR_NAME_LOCPAR(cstring));
   cOut = (unsigned char *) malloc((sizeInit +1) * sizeof(unsigned char));

   if (cOut) {
      memset(cOut,   0, sizeInit);
      memcpy(cOut, cIn, sizeIn  );
      pp = 0;                                        /* no previous byte yet */
 
      for (ptr=cOut; ptr <= (cOut + sizeIn + added / 8); ptr++, byteNo++) {
         move = 0;                         /* one or two shifts are possible */
         cc = *ptr;
         if (pp & 0x1F) {                            /* check previous byte? */
            if ((pp & 0x1F) == 0x1F) {       /* xxx11111,xxxxxxxx = 0xxxxxxx */
               cc = (cc >> 1) & 0x7F;
               move++;
            }
            else if (((pp & 0x0F) == 0x0F) && 
                     ((cc & 0x80) == 0x80)) { /* xxxx1111,1xxxxxxx = 10xxxxxx */
               cc = ((cc >> 1) & 0x3F) | 0x80;
               move++;
           }
            else if (((pp & 0x07) == 0x07) && 
                     ((cc & 0xC0) == 0xC0)) { /* xxxxx111,11xxxxxx = 110xxxxx */
               cc = ((cc >> 1) & 0x1F) | 0xC0;
               move++;
           }
            else if (((pp & 0x03) == 0x03) && 
                     ((cc & 0xE0) == 0xE0)) { /* xxxxxx11,111xxxxx = 1110xxxx */
               cc = ((cc >> 1) & 0x0F) | 0xE0;
               move++;
           }
            else if (((pp & 0x01) == 0x01) && 
                     ((cc & 0xF0) == 0xF0)) { /* xxxxxxx1,1111xxxx = 11110xxx */
               cc = ((cc >> 1) & 0x07) | 0xF0;
               move++;
            }
         }
         if (cc) {                                    /* check current byte? */
            if ((cc & 0xF8) == 0xF8) {                /* 11111xxx = 111110xx */
               cc = ((cc >> 1) & 0x03) | 0xF8;
               move++;
            }
            else if ((cc & 0x7C) == 0x7C) {           /* x11111xx = x111110x */
               dd = cc & 0xFC;
               cc = ((cc >> 1) & 0x01) | dd;
               move++;
            }
            else if ((cc & 0x3E) == 0x3E) {           /* xx11111x = xx111110 */
               cc = cc & 0xFE;
               move++;
            }
         }

         if (move) {                        
            rr = *ptr << (8 - move);                  /* remainder moved out */
            *ptr = cc;                                /* save current byte   */

            added += move;

            for (buff=ptr+1; buff <= (cOut + sizeIn + added / 8 +1); buff++) {
               dd = *buff << (8 - move);                    /* new remainder */
               *buff = (*buff >> move) | rr;   /* shift and insert remainder */
               rr = dd;
            }
         }

#ifdef TESTPRINT
 printbuff(cc, cOut, sizeInit);
#endif

         pp = cc;                                     /* next: previous byte */
      }

      sizeOut = sizeIn + added / 8;
      if ((added % 8) > 0)
         sizeOut++;
      SET_VAR_CHRLEN(VAR_NAME_LOCAL(cret), cOut, sizeOut);
      free(cOut);
   }
}
#endCinline
endif
return cRet
    
*****************************************************************************
* zeroremove(cString) -> cNewStr
*
* Removes a 0-bit after every fifth 1-bit - reverse of zeroinsert()
*
* cString : char string from zeroinsert()
* cNewStr : <cString> with removed 0-bits, the new length may be 20% shorter
*----------------------------------------------------------------------------
FUNCTION zeroremove(cString)
local cRet := ""
if valtype(cString) == "C" .and. len(cString) > 0
#Cinline
{
   int sizeIn, sizeOut, sizeInit, removed = 0, move = 0, byteNo = 0;
   unsigned char *cIn, *cOut, *ptr, *buff;
   unsigned char cc, pp, dd, rr;
   
   sizeIn   = VAR_ISLEN(VAR_NAME_LOCPAR(cstring)) -1;
   sizeInit = sizeIn +1;
   cIn  = VAR_CHR(VAR_NAME_LOCPAR(cstring));
   cOut = (unsigned char *) malloc((sizeInit +1) * sizeof(unsigned char));
   byteNo = sizeIn -1;

   if (cOut) {
      memset(cOut,   0, sizeInit);
      memcpy(cOut, cIn, sizeIn  );
 
      for (ptr=cOut + sizeIn -1; ptr >= cOut; ptr--, byteNo--) {
         move = 0;                         /* one or two shifts are possible */
         cc = *ptr;
         if (ptr > cOut)
            pp = *(ptr -1);
         else
            pp = 0;
         if (cc) {     
            if ((cc & 0xFC) == 0xF8) {                /* 111110xx = 11111xx+ */
               cc = (cc << 1) | 0xF8;
               move++;
            }
            else if ((cc & 0x7E) == 0x7C) {           /* x111110x = x11111x+ */
               dd = cc & 0xFC;
               cc = ((cc & 1) << 1) | dd;
               move++;
            }
            else if ((cc & 0x3F) == 0x3E) {           /* xx111110 = xx11111+ */
               cc = cc & 0xFE;
               move++;
            }
         }
         if (move) {                        
            rr = *(ptr +1) >> (8 - move);            /* remainder moved in  */
            cc = cc | rr;                            /* save curr result    */
         }
         if (pp & 0x1F) {     
            if (((pp & 0x1F) == 0x1F) && 
                ((cc & 0x80) == 0)) {        /* xxx11111,0xxxxxxx = xxxxxxx+ */
               cc = cc << 1;
               move++;
            }
            else if (((pp & 0x0F) == 0x0F) && 
                     ((cc & 0xC0) == 0x80)) { /* xxxx1111,10xxxxxx = 1xxxxxx+ */
               cc = (cc << 1) | 0x80;
               move++;
           }
            else if (((pp & 0x07) == 0x07) && 
                     ((cc & 0xE0) == 0xC0)) { /* xxxxx111,110xxxxx = 11xxxxx+ */
               cc = (cc << 1) | 0xC0;
               move++;
           }
            else if (((pp & 0x03) == 0x03) && 
                     ((cc & 0xF0) == 0xE0)) { /* xxxxxx11,1110xxxx = 111xxxx+ */
               cc = (cc << 1) | 0xE0;
               move++;
           }
            else if (((pp & 0x01) == 0x01) && 
                     ((cc & 0xF8) == 0xF0)) { /* xxxxxxx1,11110xxx = 1111xxx+ */
               cc = (cc << 1) | 0xF0;
               move++;
            }
         }

         if (move) {                        
            rr = *(ptr +1) >> (8 - move);            /* remainder moved in  */
            *ptr = cc | rr;                          /* save curr result    */

            removed += move;

            for (buff=ptr+1; buff <= (cOut + sizeIn - removed / 8); buff++) {
               rr = *(buff +1) >> (8 - move);
               *buff = (*buff << move) | rr;
            }
         }
      }

      sizeOut = sizeIn - removed / 8;
      if ((removed % 8) > 0)
         sizeOut--;
      SET_VAR_CHRLEN(VAR_NAME_LOCAL(cret), cOut, sizeOut);
      free(cOut);
   }
}
#endCinline
endif
return cRet

*****************************************************************************
* fs2_termio(nType, nPort, [value], [bit], [parity], [stop], 
*            [lCtsRts], [lLocal], [nCanAss]) -> ok
*
* Helper for com_hard(), com_init(), com_init2()
* assumes the port is open. The parameters depends on the 
* calling function and <nType>.
*
* nType = 0: init serial port with predefined values - for com_init()
*         1: return/set status of RTS/CTS (log)
*         2: value = baud (general, not port specific!) (num)
*         3: value = parity  (1/0)
*         4: value = dataLng (num 7/8)
*         5: value = stopBit (num 0/1/2)
*         6: return/set status of XON/XOFF (log)
*        -1: save port status    - for com_open() only
*        -2: restore port status - for com_close() only
* nPort = port (1..16)
*----------------------------------------------------------------------------
STATIC FUNCTION fs2_termio(nType, nPort, value, bit, parity, stop, ;
                           lCtsRts, lLocal, lCanon)
local RetVal := NIL
local_int iType, iFd

iType := nType
iFd   := anFileHd[nPort]

#Cinline
{
#if defined(FGSLINUX)
#   include <termios.h>
#       define FGS_termio  termios
#else
#  if defined(NSCSinix)
#    include <termios.h>
#       define FGS_termio  termios
#  else
#    include <sys/ioctl.h>
#    if defined(FGSHPUX11)
#      include <sys/termio.h>
#    else
#      if defined(FGSHPUX) || defined(FGSDECAlpha)
#         include <termio.h>
#      endif
#    endif
#    define FGS_termio  termio
#  endif
#endif

#if defined(FGSHPUX11)
#    define TCIFLUSH	0    /* from /usr/include/sys/termios.h */
#    define TCSANOW	0
#    define VSUSP	13		/* SUSP character */
#    define VSTART	14		/* START character */
#    define VSTOP	15		/* STOP character */
#endif

   struct FGS_termio term;
   struct FGS_termio *oldterm;

   int    ok, ii = 0, iVal = 0, iSpeed = 0; 
   int    iBit = 8, cParity = 'N', iStop = 1;
   int    iCtsRts = 1, iLocal = 1, iCanon = 0;

   ok = tcgetattr(ifd, &term);
   if (ok == 0) {
      if(IS_VAR_NUM(VAR_NAME_LOCPAR(value)))
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
      switch (iVal) {
        case    300: iSpeed = B300    ; break;
        case   1200: iSpeed = B1200   ; break;
        case   2400: iSpeed = B2400   ; break;
        case   4800: iSpeed = B4800   ; break;
        case   9600: iSpeed = B9600   ; break;
        case  19200: iSpeed = B19200  ; break;
        case  38400: iSpeed = B38400  ; break;
#ifdef B57600
        case  57600: iSpeed = B57600  ; break;
#endif
#ifdef B115200
        case 115200: iSpeed = B115200 ; break;
#endif
#ifdef B230400
        case 230400: iSpeed = B230400 ; break;
#endif
      }

      switch (itype) {
      case -1:
         oldterm = (struct FGS_termio *) malloc(sizeof(struct FGS_termio));
         if (oldterm) {
            memcpy(oldterm, &term, sizeof(struct FGS_termio));
            SET_VAR_SPECIAL(VAR_NAME_LOCPAR(value), oldterm);
         }
         break;
      case -2:
         oldterm = VAR_SPECIAL(VAR_NAME_LOCPAR(value));
         tcflush(ifd, TCIFLUSH);        
         ok = tcsetattr(ifd, TCSANOW, oldterm); 
         free(oldterm);
         break;   

      case 0:                
/*       bzero(&term, sizeof(term));   clear all flags */
         memset(&term, 0, sizeof(term));  /* clear all flags */

         term.c_cflag = iSpeed;  /* speed */

         if(IS_VAR_NUM(VAR_NAME_LOCPAR(bit)))
            iBit = VAR_INTNUM(VAR_NAME_LOCPAR(bit));
         switch (iBit) {
            case 5 : term.c_cflag |= CS5; break;
            case 6 : term.c_cflag |= CS6; break;
            case 7 : term.c_cflag |= CS7; break;
            case 8 : 
            default: term.c_cflag |= CS8; break;
         }
         if(IS_VAR_CHR(VAR_NAME_LOCPAR(parity)))
            cParity = *(VAR_CHR(VAR_NAME_LOCPAR(parity)));
         switch (cParity) {
            case 'O':  term.c_cflag |= PARODD;  /* fall thru */
            case 'E':  term.c_cflag |= PARENB; break; 
         }
         if(IS_VAR_NUM(VAR_NAME_LOCPAR(stop)))
            iStop = VAR_INTNUM(VAR_NAME_LOCPAR(stop));
         if (iStop == 2) 
            term.c_cflag |= CSTOPB;

         if(IS_VAR_LOG(VAR_NAME_LOCPAR(lctsrts)))
            iCtsRts = IS_VAR_TRUE(VAR_NAME_LOCPAR(lctsrts));

#ifdef CRTSCTS
         if(iCtsRts)
            term.c_cflag |= CRTSCTS;
#endif

         if(IS_VAR_LOG(VAR_NAME_LOCPAR(llocal)))
            iLocal = IS_VAR_TRUE(VAR_NAME_LOCPAR(llocal));
         if(iLocal)   
            term.c_cflag |= CLOCAL;

         term.c_cflag |= CREAD;    /* enable receiving chars */

         term.c_oflag = 0;                 /* raw output */

         if(IS_VAR_LOG(VAR_NAME_LOCPAR(lcanon)))
            iCanon = IS_VAR_TRUE(VAR_NAME_LOCPAR(lcanon));
         if(iCanon) {
            term.c_lflag = ICANON;            /* disable echo */
            term.c_iflag = IGNPAR | ICRNL;    /* ignore bytes with parity err, map CR to NL */
            term.c_cc[VINTR]    = 3;          /* Ctrl-c ENABLED */
            term.c_cc[VEOF]     = 4;          /* Ctrl-d ENABLED */
            term.c_cc[VMIN]     = 1;          /* blocking read until 1 character arrives */

            term.c_cc[VQUIT]    = 0;          /* Ctrl-\ */
            term.c_cc[VERASE]   = 0;          /* del */
            term.c_cc[VKILL]    = 0;          /* @ */
            term.c_cc[VTIME]    = 0;          /* inter-character timer unused */
#ifdef VSWTC
            term.c_cc[VSWTC]    = 0;          /* '\0' */
#endif
#ifdef VREPRINT
            term.c_cc[VREPRINT] = 0;          /* Ctrl-r */
#endif
#ifdef VDISCARD
            term.c_cc[VDISCARD] = 0;          /* Ctrl-u */
#endif
#ifdef VWERASE
            term.c_cc[VWERASE]  = 0;          /* Ctrl-w */
#endif
#ifdef VLNEXT
            term.c_cc[VLNEXT]   = 0;          /* Ctrl-v */
#endif
            term.c_cc[VSTART]   = 0;          /* Ctrl-q */
            term.c_cc[VSTOP]    = 0;          /* Ctrl-s */
            term.c_cc[VSUSP]    = 0;          /* Ctrl-z */
            term.c_cc[VEOL]     = 0;          /* '\0' */
            term.c_cc[VEOL2]    = 0;          /* '\0' */
         } else {
            term.c_lflag = 0;
            term.c_iflag = IGNPAR ;           /* ignore bytes with parity error */
            term.c_cc[VMIN]     = 1;          /* blocking read until 1 character arrives */
         }
 
         tcflush(ifd, TCIFLUSH);              /* clean modem line */
         ok = tcsetattr(ifd, TCSANOW, &term); /* set new mode */
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok == 0 ? 1 : 0);
         break;

      case 1:
         ok = 0;
#ifdef CRTSCTS
         ok = term.c_oflag & CRTSCTS;
         if(IS_VAR_LOG(VAR_NAME_LOCPAR(value))) {  /* set */
            ii = VAR_LOG(VAR_NAME_LOCPAR(value));
            if (ii) term.c_oflag |= CRTSCTS;
            else    term.c_oflag &= (~CRTSCTS);
            tcsetattr(ifd, TCSANOW, &term);
         }
#endif
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
         break;

      case 2:
         if (iSpeed) {
            ok  = cfsetispeed(&term, iSpeed);
            ok &= cfsetospeed(&term, iSpeed);
         } else
            ok = cfgetospeed(&term);
         SET_VAR_INT(VAR_NAME_LOCAL(retval),  ok);
         break;

      case 3:
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
         if (iVal) {
            term.c_iflag |= INPCK;
            term.c_oflag |= PARENB;
         } else {
            term.c_iflag |= INPCK;
            term.c_oflag &= (~PARENB);
         }
         ok = tcsetattr(ifd, TCSANOW, &term);
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), !ok);
         break;

      case 4:
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
         if (iVal == 7)
            ii = CS7;
         else if (iVal == 8 )
            ii = CS8;
         if (ii) {
            term.c_cflag &= (~CS7) & (~CS8);
            term.c_cflag |= ii;
            ok = tcsetattr(ifd, TCSANOW, &term);
         } else 
           ok = -1;
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), !ok);
         break;

      case 5:
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
         switch (iVal) {
            case 0: 
            case 1: 
               term.c_cflag &= (~CSTOPB);
               break;
            case 2: 
               term.c_cflag |= CSTOPB;
               break;
         }
         ok = tcsetattr(ifd, TCSANOW, &term);
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), !ok);
         break;

      case 6:
         ok = term.c_iflag & IXON;
         if(IS_VAR_LOG(VAR_NAME_LOCPAR(value))) {  /* set */
            ii = VAR_LOG(VAR_NAME_LOCPAR(value));
            if (ii) term.c_iflag |= IXON | IXOFF;
            else    term.c_iflag &= (~IXON) & (~IXOFF);
            tcsetattr(ifd, TCSANOW, &term);
         }
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
         break;
      }
   }
}
#endCinline
return RetVal

** eof fs2_com.prg

