****************************************************************************
*
* smallRDD.prg == "smallRDD" RDD driver    (c) 1995 by multisoft GmbH
*
* Very small RDD driver for .dbf files, demonstrating how to use OOPs and
* the DBserver RDDs. Written in the FlagShip .prg language only, does NOT
* require any C knowledge. You may freely modify it for your own purposes,
* e.g. to access SQL databases by using e.g. Inline-C while calling ESQL
* of the SQL package.
*
* Heterogenous dbf access via object and/or command + function is supported
* here. There are two FlagShip functions (RDDannAlias and RDDannField) which
* connect (announce) the Object to the FlagShip functions and working area,
* see the oRdd:INIT() method below. If used, the complementary functions
* (RDDretAlias and RDDannField) retire this connenction when the database
* gets closed (see the oRdd:CLOSE() method below).
*
* This driver uses late .dbf update, which speeds up the process, since all
* fields of the same record are updated at once. The record is writen only
* if it was changed and a movement is issued, or if the last change is older
* than 5 seconds. You may disable this "late" updating by setting the #define
* WAIT_FOR_UPDATE to 0 (zero) or set the time difference to any higher value.
*
* Note, that the default DBFIDX driver uses other variables and more
* optimized techniques than the simplified ones here. For a full DBserver
* specification, see the *.c files in /usr/FSsrc/system/rdd4cb.tar.Z file.
*
* Methods: smallRDD:INIT()      --- internal, for object instantiation only
*          smallRDD:CLOSE()     <=> USE, DBCLOSEAREA()
*          smallRDD:SKIP ([n])  <=> SKIP [n], DBSKIP([n])
*          smallRDD:GOTO (n)    <=> GO TO n,  DBGOTO(n)
*          smallRDD:FIELDGET()  <=> FIELDGET(nPos|cName)
*          smallRDD:FIELDPUT()  <=> FIELDUPT(nPos|cName, value)
*          smallRDD:DBSTRUCT()  <=> DBSTRUCT()
*          smallRDD:NOIVARGET() --- run-time access to oRdd:<field>
*          smallRDD:NOIVARPUT() --- run-time assign to oRdd:<field>
* Access:  smallRDD:USED        <=> USED()
*          smallRDD:ALIAS       <=> ALIAS()
*          smallRDD:DELETED     <=> DELETED()
*          smallRDD:RECNO       <=> RECNO()
*          smallRDD:RECCOUNT    <=> RECCOUNT(), LASTREC()
*          smallRDD:RECSIZE     <=> RECSIZE()
*          smallRDD:EOF         <=> EOF()
*          smallRDD:BOF         <=> BOF()
* Assign:  smallRDD:ALIAS       --- invoked from USE and DBUSEAREA()
* Export:  smallRDD:lEOF        --- (not required at all)
*          smallRDD:lBOF        --- (not required at all)
*
* Compile: FlagShip -na -c smallRDD.prg
* Include: rename the 'reposit.fh' produced by FScomp to 'smallRDD.fh' and
*            #include "smallRDD.fh" to your applic (myappli*.prg)
*          for prototyping, if used in your application (higly recommended).
* Link   : FlagShip myappli*.prg smallRDD.o
*
* Usage a) USE mydbf NEW VIA smallRDD
*          if USED()
*            GOTO x, SKIP, ? <field>, lEOF(), lBOF(), ...
*            REPLACE <field> ..., FIELDPUT(field,var), ...
*            USE
*          endif
*
* Usage b) #include "smallRDD.fh"          // recommended but not necessary
*          Local oDbf AS smallRDD          // recommended but not necessary
*          oDbf := SMALLRDD {"mydbf",,,,,.T.}
*          if oDbf:USED
*             oDbf:GOTO(x), oDbf:SKIP(), QOUT(oDbf:field, oDbf:lEOF, oDbf:lBOF)
*             oDbf:field := ..., oDbf:FIELDPUT(field,val), ...
*             oDbf:CLOSE()
*          endif
*
* Usage c) all combinations of (a) and (b) are supported, e.g.:
*          #include "smallRDD.fh"          // recommended
*          Local oDbf AS smallRDD          // recommended
*          USE mydbf NEW VIA smallRDD
*          if USED()
*             oDbf := DBOBJECT()
*             oDbf:GOTO(x), SKIP, ? oDbf:field, oDbf:lEOF, BOF()
*             DBCLOSEAREA()
*          endif
*
* Example: srddtest.prg
*
****************************************************************************

#include "dbstruct.fh"
#include "fileio.fh"
#include "error.fh"
#include "dataserv.fh"

* #define DEBUGGING
* #define DEBUGGING3

#define WAIT_FOR_UPDATE 5              /* set to 0 to write immediately   */
#define CONSIDER_SET_DELETED           /* SET DELETED is considered       */
#define TEST_ONLY                    /* activate for test purposes      */

#define ERR_NoDbfUsed   1
#define ERR_ReplOnEof   2
#define ERR_ReadError   3

CLASS smallRDD INHERIT DataServer
   EXPORT  lEof, lBof      AS logical    // EOF(), BOF()
   PROTECT coAlias :=""    AS character  // ALIAS()
   PROTECT loUsed  := .F.  AS logical    // USED()
   PROTECT coFullFileName  AS character  // DBF()
   PROTECT ioRecLen  := 0  AS intVar     // RECSIZE()
   PROTECT ioRecNo   := 0  AS intVar     // RECNO()
   PROTECT ioLastRec := 0  AS intVar     // LASTREC()
   PROTECT aoFields  := {} AS usual      // DBSTRUCT()

   PROTECT ioHandle  := 0  AS intVar     // handle used
   PROTECT ioStartPos:= 0  AS intVar     // relat.position of 1st record (0..)
   PROTECT ioEndPos  := 0  AS intVar     // relat.position of EOF
   PROTECT aoRelPos  := {} AS array      // relat.position of field in buff
   PROTECT coOneRecBuf     AS character  // buffer for one record
   PROTECT aoFldPtr  := {} AS array      // internal array of _FIELD vars
   PROTECT loChanged       AS logical    // is record changed?
   PROTECT noChangeTime:=0 AS numeric    // last change writen (abs.seconds)

PROTOTYPE METHOD goto()        CLASS smallRDD    // invoked before definition
PROTOTYPE METHOD close()       CLASS smallRDD    // invoked before definition

PROTOTYPE PROTECT METHOD ReadRecord()  CLASS smallRDD // invoked before def.
PROTOTYPE PROTECT METHOD WriteRecord() CLASS smallRDD // invoked before def.
PROTOTYPE PROTECT METHOD fillRddInstances() CLASS smallRDD

#ifdef TEST_ONLY
   PROTOTYPE METHOD displayInstances() CLASS smallRDD // invoked before def.
#endif

******************************************************************************
INIT procedure _smallrdd
   REQUEST smallrddNEW        // will be created automatically
   AnnNewRdd ("SmallRDD")     // in rddsys.prg, announce for RddList()
RETURN

******************************************************************************
* oRdd:INIT(cFileName, [lShare], [lReadOnly],
*           [NIL], [NIL], [lNew]) -> self               internally used Method
*
* cFilename : name of the .dbf file, optionally with path and extention
* lShare    : ignored here, allways .F. (exclusive)  used
* lReadOnly : ignored here, allways .F. (read/write) used
* lNew      : if set .F., the current working area is used. If specified
*             .T. or not given, a new working area is used. This param. is
*             equivalent to the (.not. NEW) clause of USE.
*
* Called internally on object instantiation or from USE, DBUSEAREA().
* Opens the database (always .dbf), fs_set() and SET PATH is considered.
* Fills object instances. Notifies FlagShip about the alias/work area and
* fields in use for heterogenous program usage.
*-----------------------------------------------------------------------------
METHOD init(cName, lShare, lReadonly, dummy1, dummy2, lNew) CLASS smallRDD

   Local ii, jj AS int
   if !(valtype(lNew) == "L") .or. !lNew
      SELECT 0
   else
      if loUsed .and. ioHandle > 0
         SELF:close()
      endif
   endif
   ioHandle := 0
   NETERR (.T.)                                   // set NetErr()
   if valtype(cName) != "C" .or. empty(cName)     // param checking
      loUsed := .F.
      return self
   endif
   loUsed := rddOpenDBF (@cName, @ioHandle)       // see Fn() below
   if loUsed
      loUsed := SELF:fillRddInstances (cName)     // see method below
   endif
   if loUsed
      coAlias := RDDannAlias (self,             ; // returns actual
                              coFullFileName)     // alias name
      if empty (coAlias)
         alert ("smallRDD:INIT() cannot create/modify alias '" + ;
                coAlias + "'")
         SELF:close()
         return self
      endif
      jj := len(aoFields)
      aoFldPtr := ARRAY(jj)                       // create carr.array
      for ii := 1 to jj                           // Announce _FIELD Vars:
         RDDannField (aoFields[ii,DBS_NAME],    ; // -field name
                      ii,                       ; // -field pos
                      aoFields[ii,DBS_TYPE],    ; // -field type
                      1,                        ; // -late field fill
                      aoFields[ii,DBS_LEN ],    ; // -field length
                      aoFields[ii,DBS_DEC ],    ; // -field deci length
                      aoFldPtr)                   // -carrying array
      next
      coOneRecBuf := space(ioRecLen)              // assign buffer
      SELF:goto(1)                                // read first record
      NETERR (.F.)                                // reset NetErr()
   endif
   noChangeTime:= 0                               // init instance
   loChanged   := .F.                             // init instance
#ifdef TEST_ONLY
   self:displayInstances(.T.)
#endif
   return self

******************************************************************************
* oRdd:CLOSE() -> lSuccess                                              Method
*
* Close the database, free the assigned work area, alias and fields
* Equivalent to USE or DBCLOSEAREA()
*-----------------------------------------------------------------------------
METHOD close() CLASS smallRDD
   Local ok := .F., ii
   if loUsed .and. ioHandle > 0                   // is the .dbf open ?
      if loChanged                                // update record if changed
         SELF:WriteRecord(.T.)
      endif
      FSEEK  (ioHandle, 0, FS_START)              // try to unlock
      FLOCKF (ioHandle, FF_UNLOCK)                // the whole .dbf
      FSEEK  (ioHandle, 0, FS_END)                // required for some systems
      ok := FCLOSE (ioHandle)                     // close
   endif
   for ii := 1 to len(aoFldPtr)                   // free _FIELD vars
      RDDretField (self, aoFields[ii,DBS_NAME])   // in FlagShip
   next
   coAlias       := ""                            // reset instances
   aoFields      := {}
   loUsed        := .F.
   ioHandle      := 0
   RDDretAlias (self)                             // free Alias in FS
   return ok

******************************************************************************
* oRdd:USED -> lUsed                                                    Access
* equivalent to USED()
*-----------------------------------------------------------------------------
ACCESS used CLASS smallRDD
   return loUsed

******************************************************************************
* oRdd:lEOF  -> lEndOfFile                                              Access
* equivalent to EOF()
*-----------------------------------------------------------------------------
ACCESS eof CLASS smallRDD
   return lEof

******************************************************************************
* oRdd:lBOF  -> lBeginOfFile                                            Access
* equivalent to BOF()
*-----------------------------------------------------------------------------
ACCESS bof CLASS smallRDD
   return self:lBof

******************************************************************************
* oRdd:RECCOUNT -> nRecords                                             Access
* equivalent to RECCOUNT(), LASTREC()
*-----------------------------------------------------------------------------
ACCESS reccount CLASS smallRDD
   return INT2NUM(ioLastRec)        // for ValType() == 'N'

******************************************************************************
* oRdd:RECSIZE  -> nRecordSize                                          Access
* equivalent to RECSIZE()
*-----------------------------------------------------------------------------
ACCESS recsize CLASS smallRDD
   return INT2NUM(ioRecLen)        // for ValType() == 'N'

******************************************************************************
* oRdd:RECNO <-> nRecno                                          Access/Assign
* equivalent to RECNO() or DBGOTO(nRecno)
*-----------------------------------------------------------------------------
ACCESS recno CLASS smallRDD
   return INT2NUM(ioRecNo)        // for ValType() == 'N'

ASSIGN recno(value) CLASS smallRDD
   if valtype(value) $ "NI"
      self:GOTO(value)
   endif
   return INT2NUM(ioRecNo)        // for ValType() == 'N'

******************************************************************************
* oRdd:ALIAS <-> cAlias                                          Access/Assign
*
* equivalent to ALIAS().
* Caution: do NOT change the alias when using hybrid programming!
*-----------------------------------------------------------------------------
ACCESS alias CLASS smallRDD
   return coAlias

ASSIGN alias(value) CLASS smallRDD
   LOCAL cAlias

   if valtype(value) == "C"
      cAlias := RDDannAlias (self, value)        // returns actual alias name
      if empty (cAlias)
         alert ("smallRDD:ALIAS() cannot assign alias '" + cAlias + "'")
         return coAlias
      endif
     coAlias := cAlias
   endif
   return coAlias

******************************************************************************
* oRdd:DELETED -> lDeleted                                              Access
* equivalent to DELETED()
*-----------------------------------------------------------------------------
ACCESS deleted CLASS smallRDD
   return left(@coOneRecBuf,1) $ "*TtYy"

*-----------------------------------------------------------------------------
PROTECT METHOD preSkipCheck() CLASS smallRDD                       // internal
   if EMPTY(loUsed)
      displErr (ERR_NoDbfUsed)
      return .F.
   endif
   if ioLastRec == 0
      lbof := lEof := .T.
      return .F.
   endif
   return .T.

*-----------------------------------------------------------------------------
PROTECT METHOD postSkipCheck(new AS intVar) CLASS smallRDD         // internal
   if new < ioStartPos .or. ioRecNo < 0
      FSEEK (ioHandle, ioStartPos, FS_START)
      lBof     := .T.
      ioRecNo := 1
      SELF:ReadRecord()                               // read record in buff
      return .F.
   endif
   if new >= ioEndPos .or. ioRecNo > ioLastRec
      lEof     := .T.
      ioRecNo := ioLastRec +1
      SELF:ReadRecord()                               // read record in buff
      return .F.
   endif
   return .T.

*-----------------------------------------------------------------------------
PROTECT METHOD getFieldValue (nFieldNo, type)  CLASS smallRDD      // internal
   DO CASE
   CASE type == "C"
      return substr(coOneRecBuf, aoRelPos[nFieldNo],aoFields[nFieldNo,DBS_LEN])
   CASE type == "L"
      return (substr(coOneRecBuf, aoRelPos[nFieldNo], 1) $ "TtYy")
   CASE type == "N"
      return val(substr(coOneRecBuf, aoRelPos[nFieldNo], ;
                 aoFields[nFieldNo,DBS_LEN]), aoFields[nFieldNo,DBS_DEC])
   CASE type == "D"
      return stod(substr(coOneRecBuf, aoRelPos[nFieldNo], 8))
   OTHERWISE
      return NIL
   ENDCASE

*-----------------------------------------------------------------------------
PROTECT METHOD putFieldValue (nFieldNo, value)  CLASS smallRDD     // internal
   Local cType := valtype(value), cBuff
   if nFieldNo <= 0 .or. nFieldNo > len(aoFields) .or. ;
      !(cType == aoFields[nFieldNo,DBS_TYPE])
      return .F.                                      // wrong type
   endif
   DO CASE
   CASE ctype == "C"
      cBuff := PADR (value, aoFields[nFieldNo,DBS_LEN])
   CASE ctype == "L"
      cBuff := if (value, "T", "F")
   CASE ctype == "N"
      cBuff := str(value, aoFields[nFieldNo,DBS_LEN], ;
                   aoFields[nFieldNo,DBS_DEC])
   CASE ctype == "D"
      cBuff := dtos (value)
   ENDCASE
   if cbuff != NIL                                    // supported type?
      stuff (coOneRecBuf, aoRelPos[nFieldNo], ;
             aoFields[nFieldNo,DBS_LEN], cBuff)         // replace record buff
      loChanged := .T.
      return .T.
   endif
   return .F.

*-----------------------------------------------------------------------------
PROTECT METHOD preGetCheck(cVar, cText1, cText2) CLASS smallRDD    // internal
   Local VarPos
   VarPos := ASCAN (aoFields, {|xarr| upper(cVar) == trim(xarr[DBS_NAME]) })
   if VarPos == 0
      if !empty(cText1)
         ALERT ("Unknown " + cText1 + " '" + cVar + ;
                "';of the class smallRDD " + cText2 + " in;" + ;
                procname(2) + "/" + alltrim(str(procline(2))))
      endif
   endif
   return VarPos

******************************************************************************
* oRdd:FIELDGET(nFieldPos|cFieldName) -> value                          Method
*
* Equivalent to FIELDGET() function.
* This method is also automatically invoked on <value> := _FIELD-><name>
* according to the RDDannField() specification in oRdd:init()
*-----------------------------------------------------------------------------
METHOD fieldget(ncField) CLASS smallRDD
   Local nFieldNo := 0, cFieldName, value, cType
   if EMPTY(loUsed)
#ifdef DEBUGGING1
   ? "exit fg ! loused", loused
#endif
      return NIL
   endif
   if valtype(ncField) == "C"
      cFieldName := left(upper(alltrim(ncField)), 10)
      nFieldNo := SELF:preGetCheck (cFieldName, "Field", ;
                                    "accessed by FIELDGET()")
   elseif valtype(ncField) $ "NI"
      nFieldNo := ncField
   else
#ifdef DEBUGGING1
   ? "exit ncfield", ncfield, valtype(ncfield)
#endif
      return NIL                                      // wrong param
   endif
   if nFieldNo <= 0 .or. nFieldNo > len(aoFldPtr)
#ifdef DEBUGGING1
   ? "exit fieldno o.f.range", nfieldno, len(aofldptr)
#endif
      return NIL
   endif
   ctype := aoFields[nFieldNo,DBS_TYPE]
   value := SELF:getFieldValue (nFieldNo, cType)      // field -> variable
   if value == NIL
      ALERT ("Unknown/unsupported type '" + ctype + "' of Field" + ;
             if (valtype(ncField) == "C", "-Name '" + cFieldName + "'", ;
                 "-Number " + ltrim(str(nFieldNo)) ) + ;
             ";of the class smallRDD accessed in;"   + ;
             procname(1) + "/" + alltrim(str(procline(1))))
      return NIL
   endif
   AssignFldValue (@aoFldPtr[nFieldNo], value)        // don't change SubType
   return @aoFldPtr[nFieldNo]                         // varSubType = field

******************************************************************************
* oRdd:FIELDPUT(nFieldPos|cFieldName, value) -> lSuccess                Method
*
* Equivalent to FIELDPUT() function.
* This method is also automatically invoked on _FIELD-><name> := <value>
* according to the RDDannField() specification in oRdd:init()
*-----------------------------------------------------------------------------
METHOD fieldput(ncField, value) CLASS smallRDD
   Local nFieldNo := 0, cFieldName, cType, ok

   if EMPTY(loUsed) .or. SELF:lEof
      displErr (if (EMPTY(loUsed), ERR_NoDbfUsed, ERR_ReplOnEof))
      return .F.
   endif
   if valtype(ncField) == "C"
      cFieldName := left(upper(alltrim(ncField)), 10)
      nFieldNo := SELF:preGetCheck (cFieldName, "Field", ;
                                    "accessed by FIELDPUT()")
   elseif valtype(ncField) $ "NI"
      nFieldNo := ncField
   else
      return .F.                                      // wrong param
   endif
   if nFieldNo <= 0 .or. nFieldNo > len(aoFldPtr)
      return .F.
   endif
   ctype := aoFields[nFieldNo,DBS_TYPE]
   ok := SELF:putFieldValue (nFieldNo, value)         // variable -> fieldbuff
   if !ok
      ALERT ("Unknown/unsupported/different type '" + ctype + "' of Field" + ;
             if (valtype(ncField) == "C", "-Name '" + cFieldName + "'", ;
                 "-Number " + ltrim(str(nFieldNo)) ) + ;
             ";replaced by variable type '" + valtype(value) + "'" + ;
             ";of the class smallRDD accessed in;"   + ;
             procname(1) + "/" + alltrim(str(procline(1))))
      return .F.
   endif
   if cType == "C"
      AssignFldValue (@aoFldPtr[nFieldNo],            // assign to _FIELD->..
                      PADR (value, aoFields[nFieldNo,DBS_LEN]))
   else                                               // but don't change
      AssignFldValue (@aoFldPtr[nFieldNo], value)     // the var's SubType
   endif
   loChanged := .T.                                   // record is modified
   SELF:WriteRecord(.F.)                              // write on request
   return .T.

******************************************************************************
* oRdd:NoIvarGet(cVarName) -> value                       internal used Method
*
* called from the FS runtime system on unknown oRdd:<name> access
*-----------------------------------------------------------------------------
METHOD NoIvarGet(cVar) CLASS smallRDD
   Local nField
   nField := SELF:preGetCheck (cVar, "Field/Variable", "accessed")
   if nField == 0
      return .F.
   endif
   return @ SELF:FieldGet(nField)

******************************************************************************
* oRdd:NoIvarPut(cVarName, value) -> lSuccess             internal used Method
*
* called from the FS runtime system on unknown oRdd:<name> assign
*-----------------------------------------------------------------------------
METHOD NoIvarPut(cVar, value) CLASS smallRDD
   Local nField
   nField := SELF:preGetCheck (cVar, "Field/Variable", "assigned")
   if nField == 0
      return .F.
   endif
   return SELF:FieldPut(nField, @value)


******************************************************************************
* oRdd:SKIP([nNoOfRecords]) -> lSuccess                                 Method
*
* Equivalent to SKIP [n] command and DBSKIP([n]) function
* Note: SET DELETED is considered, but not SET FILTER
*-----------------------------------------------------------------------------
METHOD skip(nCount) CLASS smallRDD
   Local iSign, new           AS IntVar
   Local iSoll, iSkipped := 0 AS INT

   if !(valtype(nCount) $ "NIU") .or. ;
      ! SELF:preSkipCheck()                           // param, used, empty ?
      RETURN .F.
   endif
   if loChanged                                       // curr.record changed?
      SELF:WriteRecord(.T.)                           // yes, write first
   endif
   lBof := lEof := .F.
   nCount := default (nCount, 1)

#ifdef CONSIDER_SET_DELETED
   if !SET(_SET_DELETED)                              // SET DELETED OFF ?
#endif
#ifdef DEBUGGING3
   ? "ncount, ioreclen", ncount, ioreclen
#endif
      new := FSEEK (ioHandle, (nCount * ioRecLen), FS_RELATIVE)
      ioRecNo := ioRecNo + nCount
      SELF:ReadRecord()                               // read record in buff
      if ! SELF:postSkipCheck (new)                   // set EOF or BOF ?
         return .F.
      endif
#ifdef CONSIDER_SET_DELETED
   else                                               // SET DELETED ON
      iSign := NUM2INT(ncount / abs(ncount))          // = +/- 1
      iSoll := NUM2INT(abs(nCount))
      while !lEof .and. !lBof .and. iSkipped < iSoll
         new := FSEEK (ioHandle, (iSign * ioRecLen),
                       FS_RELATIVE)                   // skip +/- 1 record
         ioRecNo := ioRecNo + iSign
         SELF:ReadRecord()                            // read record in buff
         SELF:postSkipCheck (new)                     // set EOF or BOF
         if ! (self:Deleted .or. SET(_SET_DELETED))
            iSkipped++                                // count not deleted
         endif                                        // records only
      enddo
      if lBof .and. self:Deleted .and. SET(_SET_DELETED)
         while self:Deleted .and. !lEof
            new := FSEEK (ioHandle, ioRecLen, FS_RELATIVE)
            ioRecNo++
            SELF:ReadRecord()                         // read record in buff
            SELF:postSkipCheck (new)
         enddo
      endif
      return (iSkipped == iSoll)
   endif
#endif
   return .T.

******************************************************************************
* oRdd:GOTO (nRecNo) -> lSuccess                                        Method
*
* equivalent to GO TO n command and DBGOTO(n) function
*-----------------------------------------------------------------------------
METHOD goto(nRec) CLASS smallRDD
   Local new AS intVar
   if !(valtype(nRec) $ "NI") .or. ;
      nRec < 0 .or. nRec > (ioLastRec +1) .or. ;
      ! SELF:preSkipCheck()                           // param, used, empty ?
#ifdef DEBUGGING2
   ? "goto short out, nrec = ", nrec, valtype(nrec), "lastrec=", ;
      iolastrec, "precheck=", self:preskipcheck()
#endif
      RETURN .F.
   endif
   if loChanged                                       // curr.record changed?
      SELF:WriteRecord(.T.)                           // yes, write first
   endif
   new := FSEEK (ioHandle, ioStartPos + ((nRec -1) * ioRecLen), FS_START)
   lBof := lEof := .F.                                // reset EOF and BOF
   ioRecNo := nRec
   SELF:ReadRecord()                                  // read record in buff
   if ! SELF:postSkipCheck (new)                      // set EOF or BOF ?
      return .F.
   endif
   return .T.

******************************************************************************
* oRdd:DBSTRUCT() -> array                                              Method
*
* equivalent to DBSTRUCT() function
*-----------------------------------------------------------------------------
METHOD DBstruct() CLASS smallRDD
   return aoFields

*-----------------------------------------------------------------------------
STATIC function default (var, value)                               // internal
   return if (var == NIL, value, var)

*-----------------------------------------------------------------------------
STATIC function displErr (nMsg, var1, var2)                        // internal
   Local cWA := ltrim(str(SELECT()))
   do case
   case nMsg == ERR_NoDbfUsed
      ALERT ("No database open in WA " + cWA )
   case nMsg == ERR_ReplOnEof
      ALERT ("Cannot replace on EOF() in WA " + cWA)
   case nMsg == ERR_ReadError
      ALERT ("Read error " + ltrim(str(FERROR())) + ;
             ": in WA " + cWA + ". Only " + ltrim(str(var1)) + ;
             " instead of " + ltrim(str(var2)) + " bytes read")
   endcase
   return NIL

*-----------------------------------------------------------------------------
STATIC function rddOpenDBF (cName, iHandle)                        // internal
   Local ok := .F., cFullName := cName, iErr
   iHandle := 0
   if !(upper(right(cName, 4)) == ".DBF")
      cFullName += ".dbf"                        // add default extention
   endi
   if !file(cFullName)
      cFullName := TruePath (cFullName)          // incl.FS_SET() and SET PATH
   endif
   if file(cFullName)
      iHandle := FOPEN (cFullName, FO_READWRITE) // return handle number
   endif
   if iHandle <= 0                               // success ?
      iErr := FERROR()
      iErr := if (iErr == 0, EX_ENOENT, iErr)
      ALERT ("Database '" + cFullName + "' not found,;ErrorCode = " + ;
             ltrim(str(iErr)))
      return .F.
   endif
   cName := cFullName                            // return full file name
   return .T.

*-----------------------------------------------------------------------------
PROTECT METHOD ReadRecord() CLASS smallRDD                         // internal
   Local iCurrPos, iRead

   if SELF:lEof
      coOneRecBuf := space(ioRecLen)
#ifdef DEBUGGING2
   ? "on eof", leof
#endif
      return NIL
   endif
   iCurrPos := FSEEK (ioHandle, 0, FS_RELATIVE)        // current file posit
   if iCurrPos >= ioEndPos
      lEof := .T.
      coOneRecBuf := space(ioRecLen)
      return NIL
   endif
   iRead := FREAD (ioHandle, @coOneRecBuf, ioRecLen)   // read record
   FSEEK (ioHandle, iCurrPos, FS_START)                // go back
#ifdef DEBUGGING2
   ? "seek on", icurrpos, "read in", iread, "to read", ioreclen
#endif
   if iRead != ioRecLen
      displErr (ERR_ReadError, iRead, ioRecLen)
      coOneRecBuf := space(ioRecLen)
   endif
   return NIL

*-----------------------------------------------------------------------------
#define OFFSET_DBF_VERS 0                /* offset of .dbf database type    */
#define OFFSET_UPD_DATE 1                /* LUPDATE() offset in .dbf header */
#define OFFSET_RECCOUNT 4                /* offset of record count          */
#define OFFSET_HDRLEN   8                /* offset of header size           */
#define OFFSET_RECLEN   10               /* offset of record size           */
#define OFFSET_1STFIELD 32               /* offset of 1st field descriptor  */
#define SIZE_HEADFIELD  32               /* size of field descr in header   */

*-----------------------------------------------------------------------------
PROTECT METHOD WriteRecord(lMust) CLASS smallRDD                   // internal
   Local iCurrPos, currSeconds := seconds()
   Local cUpdDate

   if !lMust                                           // write required ?
      if !loChanged .or. WAIT_FOR_UPDATE <= 0          // no, changes pending?
         return NIL                                    // no, do nothing
      endif
      if noChangeTime > 0 .and. ;
         noChangeTime <= currSeconds .and. ;
         (currSeconds-noChangeTime) < WAIT_FOR_UPDATE  // don't write, if
         return NIL                                    // the last update
      endif                                            // was within 5 sec
   endif
   iCurrPos := FSEEK (ioHandle, 0, FS_RELATIVE)        // current file posit
   FWRITE (ioHandle, coOneRecBuf, ioRecLen)            // write record
   cUpdDate := chr(year(date() - 1900)) + ;
               chr(month(date())) + chr(day(date()))
   FSEEK  (ioHandle, OFFSET_UPD_DATE, FS_START)        // update the .dbf
   FWRITE (ioHandle, cUpdDate, 3)                      // modification date
   FSEEK  (ioHandle, iCurrPos, FS_START)               // go back
   loChanged    := .F.                                 // no changes pending
   noChangeTime := currSeconds                         // time of last update
   return NIL

*-----------------------------------------------------------------------------
PROTECT METHOD fillRddInstances(cName) CLASS smallRDD              // internal
   Local iPos, iFld, iFieldCount := 0 AS IntVar
   Local ii, iBuff, iRecPos           AS LONG
   Local cBuff := space(OFFSET_1STFIELD)

   self:lEof := self:lBof := .T.                 // set EOF() and BOF()
   coFullFileName := cName                       // set FileName instance
   cName := strtran(cName, "\", "/")             // normalize file name

   FSEEK (ioHandle, 0, FS_START)                 // read the first 16
   FREAD (ioHandle, @cBuff, 16)                  // bytes of .dbf header
   ii := strpeek(@cbuff, OFFSET_DBF_VERS +1)
   if ii != 3                                    // 0x03 = regular .dbf
      if ii == 131                               // 0x83 = .dbf and .dbt
         ALERT ("Warning: .dbt and memo fields not supported here")
      else
         ALERT ("Error: '" + coFullFileName + ;
                "' is not a regular .dbf database")
         FCLOSE (ioHandle)                       // close database
         loUsed := .F.
         return .F.
      endif
   endif
   ioRecLen := BIN2I(substr(@cbuff, OFFSET_RECLEN +1, 2))     // incl.deleted,
   if ioRecLen < 0                                            // in bytes
      ioRecLen = 65535 + ioRecLen                             // correct, if
   endif                                                      // signed
   ioLastRec  := BIN2L(substr(@cbuff, OFFSET_RECCOUNT +1, 4)) // 0..n records
   ioStartPos := BIN2I(substr(@cbuff, OFFSET_HDRLEN   +1, 2)) // 1st valid
   ioEndPos   := ioStartPos + (ioLastRec * ioRecLen)          // last valid

   FSEEK (ioHandle, 0, FS_START)                              // try flock
   ii := FLOCKF (ioHandle, FF_READWRITE, 2147483648)          // 2 GBytes
   if ii != 0                                                 // or at least
      ii := FLOCKF (ioHandle, FF_READWRITE)                   // the database
   endif
   if ii != 0
      if ALERT ("Exclusive lock on '" + cName + "' failed.;Continue ?", ;
                {"No", "Yes"}) != 2
         FCLOSE (ioHandle)                                    // close dbf
         loUsed := .F.
         return .F.
      endif
   endif

   iFld := NUM2INT((ioStartPos -1 - OFFSET_1STFIELD) / ;
                    SIZE_HEADFIELD)                           // no of fields
   aoFields := array(iFld)                                    // dbstruct()
   aoRelPos := array(iFld)                                    // rel.posit.fld
   iBuff := (iFld * SIZE_HEADFIELD) +1                        // incl eohead
   cBuff := space(iBuff)
   FSEEK (ioHandle, OFFSET_1STFIELD, FS_START)                // read field
   FREAD (ioHandle, @cBuff, iBuff)                            // descriptors
#ifdef DEBUGGING
   ? "Buffer read", len(cBuff), "ferror()", ferror()
#endif
*
*  FlagShip uses the last byte in header for consistency check
*
*  if strpeek(@cBuff, iBuff) # 13
*     ALERT ("Warning: possible .dbf header corruption")      // 0x0d = eohead
*  endif

   iPos    := 1                           // relat.position in header
   iRecPos := 2                           // since coOneRecBuf[1]=deleted flag
   for ii := 1 to iFld
#ifdef DEBUGGING
      ? "Field", ii,   upper(substr(@cBuff, ipos,   10)), ;   // fld name
                       upper(substr(@cBuff, ipos+11, 1)), ;   // fld type
                       strpeek(@cBuff, ipos+16),          ;   // fld len
                       strpeek(@cBuff, ipos+17)               // fld deci
#endif
      aoFields[ii] := {upper(substr(@cBuff, ipos,   10)), ;   // fld name
                       upper(substr(@cBuff, ipos+11, 1)), ;   // fld type
                       strpeek(@cBuff, ipos+16),          ;   // fld len
                       strpeek(@cBuff, ipos+17) }             // fld deci
      if aoFields[ii,2] == "C" .and. aoFields[ii,4] > 0       // correct the
         aoFields[ii,3] := aoFields[ii,4] * 256           ;   // char size
                         + aoFields[ii,3]                     // 1..65534
         aoFields[ii,4] := 0
      endif
      aoRelPos[ii] := iRecPos                                 // rel.fld posit
*     iRecPos += aoFields[ii,3]                               // and increment
      iRecPos = iRecPos + aoFields[ii,3]                      // TEST <<<
#ifdef DEBUGGING2
      ? "RelPos", substr(@cbuff, iPos+12, 4), aoRelPos[ii]
#endif
      iPos += SIZE_HEADFIELD
   next
   return .T.

*-----------------------------------------------------------------------------
#ifdef TEST_ONLY
METHOD displayInstances(lPrint) CLASS smallRDD
   Local lPrintOnly := if (valtype(lPrint) == "L", lPrint, .F.)
   if lPrintOnly
      Local isprint := SET (_SET_PRINTER)
      Local iscons  := SET (_SET_CONSOLE)
      SET PRINTER ON
      SET CONSOLE OFF
   endif
   ? "---- smallRDD instances at " + procname(1) +"/"+ ltrim(str(procline(1)))
   ? "loUsed,    oRdd:Used   ", loUsed, self:Used
   ? "[coAlias], oRdd:Alias  ", "["+coAlias+"]", self:Alias
   ? "eof, bof               ", lEof, lBof
   if loUsed
      ? "[coFullFileName]       ", "[" + coFullFileName + "]"
      if len(coOneRecBuf) <= 55
         ? "[coOneRecBuf]          ", "[" + coOneRecBuf + "]"
      else
         ? "[coOneRecBuf]          ", "[" + left(coOneRecBuf,44) + ;
                                      "]...[" + right(coOneRecBuf,5) + "]"
      endif
      ? "ioRecLen, oRdd:RECSIZE ", ioRecLen,  self:Recsize
      ? "ioRecNo,  oRdd:RECNO   ", ioRecNo,   self:Recno
      ? "ioLastRec,oRdd:RECCOUNT", ioLastRec, self:RECCOUNT
      ? "ioHandle               ", ioHandle
      ? "ioStartPos, ioEndPos   ", ioStartPos, ioEndPos
      ? "noChangeTime, loChanged", noChangeTime, loChanged

      ? "Dbstruct:"
      aeval (aoFields, {|x| qout(padr("["+x[1]+"]",12), x[2], x[3], x[4])})

      ? "aoFields:"
      aeval (aoRelPos, {|x| qout(x)})
      if !lPrintOnly
         wait "---ready, press any key: "
      endif
   endif
   if lPrintOnly
      SET PRINTER (isprint)
      SET CONSOLE (iscons )
   endif
   return loUsed
#endif

*** eof smallRDD.prg ***

